*-----------------------------------------------------------------------*
*	Falcon 68030 (clipped & masked) truecolour scaled sprite code,	*
*	ideal for Doom style or other 3D Falcon games.			* 
*-----------------------------------------------------------------------*
*	(c) Douglas Little 1995						*
*-----------------------------------------------------------------------*
*	Transparent pixel key is black (hex $0000)			*
*-----------------------------------------------------------------------*
*	Use the mouse to move object around - holding down left mouse	*
*	button allows x & y rescaling. No limits on object size.	*
*-----------------------------------------------------------------------*
*	Graphics are available from SCALER.ZIP found on NeST		*
*-----------------------------------------------------------------------*
*	No fudges or tricks used here - this is the simplest approach.	*
*	Both the inner & outer rendering loops of this routine fit	*
*	inside the 256-byte instruction cache with room to spare.	*
*	Hopefully this will help dispel the myth about truecolour	* 
*	being dead slow on the Falcon - it is usually dodgey		*
*	software that's to blame. 					*
*-----------------------------------------------------------------------*
*	One final word...						*
*				Long live assembly language!    ~:)	*
*-----------------------------------------------------------------------*

	OPT	O-,W-

*---------------------------->   Equates     <--------------------------*

		rsreset
bpl1		rs.b	1
bpl2		rs.b	1
bpl4		rs.b	1
bpl8		rs.b	1

		rsreset
plane_bits	rs.b	2
true_bit	rs.b	1
hires_bit	rs.b	1
vga_bit		rs.b	1
pal_bit		rs.b	1
os_bit		rs.b	1
compat_bit	rs.b	1
lace_bit	rs.b	1

true		=	1<<true_bit
hires		=	1<<hires_bit
vga		=	1<<vga_bit
pal		=	1<<pal_bit
os		=	1<<os_bit
compat		=	1<<compat_bit
lace		=	1<<lace_bit

*---------------------------->   Screen     <---------------------------*

vgaswitch		; enable for VGA / comment out for RGB/TV
;method1		; enable for render method1 (bitwise logic)
method2			; enable for render method2 (conditional logic)


		ifd	vgaswitch

video_mode:	=	vga|lace|true
screen_width:	=	320
screen_height:	=	240

		elseif

video_mode:	=	true
screen_width:	=	320
screen_height:	=	200

		endc

*---------------------------->   Macros     <---------------------------*

push	macro
	move.\0		\1,-(sp)
	endm
	
pop	macro
	move.\0		(sp)+,\1
	endm

pushall	macro
	movem.l		d0-d7/a0-a6,-(sp)
	endm
	
popall	macro
	movem.l		(sp)+,d0-d7/a0-a6
	endm

*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*
*	Program starts here			*
*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*

start:	move.l		sp,a5
	lea		ustack,sp
	move.l		$04(a5),a5
	move.l		$0C(a5),d0
	add.l		$14(a5),d0
	add.l		$1C(a5),d0
	add.l		#$100,d0
	push.l		d0
	push.l		a5
	push.w		d0
	push.w		#$4A
	trap		#1
	lea		12(sp),sp
*-----------------------------------------------*
	dc.w		$A009
*-----------------------------------------------*
*	Init display & mouse			*
*-----------------------------------------------*
	bsr		set_screen
	bsr		cls
	jsr		set_mouse
*-----------------------------------------------*
*	Set up sprite variables			*
*-----------------------------------------------*
	move.w		#0,dxpos		; x position on screen
	move.w		#0,dypos		; y position on screen
	move.w		#40,dxsize		; destination sprite width
	move.w		#60,dysize		; destination sprite height
	move.w		#40,sxsize		; source sprite width (area to use)
	move.w		#60,sysize		; source sprite height
	move.w		#screen_width,scrxsize	; screen width
	move.w		#screen_height,scrysize	; screen height
	move.w		#WID64,srcpwid		; sourcedata linewidth (actual pixel width)
	move.w		#screen_width,dstpwid	; destination linewidth (actual pixel width)
*-----------------------------------------------*
*	Draw loop				*
*-----------------------------------------------*
redraw:
*-----------------------------------------------*
*	Call sprite code			*
*-----------------------------------------------*
	lea		data+20,a4		; source = (Apex block file + 20 byte header)
	move.l		screen,a5		; dest = screen
	bsr		rasterizer
*-----------------------------------------------*
*	Loop if no keypress			*
*-----------------------------------------------*
	push.w		#11
	trap		#1
	addq		#2,sp
	tst.w		d0
	beq.s		redraw
*-----------------------------------------------*
*	Tidy up					*
*-----------------------------------------------*
exit:	jsr		reset_mouse
	bsr		reset_screen
*-----------------------------------------------*
	dc.w		$A00A
*-----------------------------------------------*
*	Terminate				*
*-----------------------------------------------*
	move.w		#0,-(sp)
	trap		#1
*-----------------------------------------------*

*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*
*	Sprite rasterizer (with clipping)	*
*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*

*-----------------------------------------------*
*	Coded binary source pixel widths	*
*-----------------------------------------------*

		rsreset
WID1		rs.b	1
WID2		rs.b	1
WID4		rs.b	1
WID8		rs.b	1
WID16		rs.b	1
WID32		rs.b	1
WID64		rs.b	1
WID128		rs.b	1
WID256		rs.b	1
WID512		rs.b	1

*-----------------------------------------------*
*	Local variable structure used by loop	*
*-----------------------------------------------*

		rsreset
cxsize:		rs.w	1
cysize:		rs.w	1
sxpos:		rs.w	1
sxposf:		rs.w	1
sypos:		rs.w	1
syposf:		rs.w	1
sxinc:		rs.w	1
sxincf:		rs.w	1
syinc:		rs.w	1
syincf:		rs.w	1
scsp_size:	rs.b	0

*-----------------------------------------------*
*	Pixel rendering unit 			*
*-----------------------------------------------*

render	macro
*-----------------------------------------------*
*	This unit uses bitwise masking		*
*-----------------------------------------------*
	ifd		method1
	move.w		(a0,d0.w*2),d3
	add.w		a2,d1
	addx.w		d2,d0
	move.l		d3,d5
	neg.l		d5
	not.l		d5
	swap		d5
	and.w		(a1),d5
	or.w		d3,d5
	move.w		d5,(a1)+
	endc
*-----------------------------------------------*
*	This unit uses conditional masking	*
*-----------------------------------------------*
	ifd		method2
	move.w		(a0,d0.w*2),d3
	beq.s		.z\@
	move.w		d3,(a1)
.z\@:	add.w		a2,d1
	addx.w		d2,d0
	addq.l		#2,a1
	endc
	endm

*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*
*	Render scaled sprite onto screen	*
*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*
rasterizer:
*-----------------------------------------------*
*	Allocate local structure on stack	*
*-----------------------------------------------*
	lea		-scsp_size(sp),sp
*-----------------------------------------------*
*	Calculate fractional increments		*
*-----------------------------------------------*
.scale:	move.w		dxsize,d5
	move.w		sxsize,d2
	ext.l		d5
	swap		d2
	clr.w		d2
	divu.l		d5,d2
*-----------------------------------------------*
	move.w		dysize,d6
	move.w		sysize,d4
	ext.l		d6
	swap		d4
	clr.w		d4
	divu.l		d6,d4
*-----------------------------------------------*
*	Calculate start positions		*
*-----------------------------------------------*
*	Initial (unclipped) values		*
*-----------------------------------------------*
.clip:	moveq		#0,d0
	moveq		#0,d1
	move.w		dxsize,d7
	move.w		dysize,d6
	move.w		dxpos,d3
	move.w		dypos,d5
*-----------------------------------------------*
*	Clip left & top edges			*
*-----------------------------------------------*
	ext.l		d3
	bpl.s		.xpos
	add.w		d3,d7
	ble		.err
	move.l		d3,d0
	muls.l		d2,d0
	neg.l		d0
	moveq		#0,d3
.xpos:	ext.l		d5
	bpl.s		.ypos
	add.w		d5,d6
	ble		.err
	move.l		d5,d1
	muls.l		d4,d1
	neg.l		d1
	moveq		#0,d5
*-----------------------------------------------*
*	X & Y source increments now ready	* 
*-----------------------------------------------*
.ypos:	move.l		d2,sxinc(sp)
	move.l		d4,syinc(sp)
*-----------------------------------------------*
*	Initial X & Y source points now ready	* 
*-----------------------------------------------*
	move.l		d0,sxpos(sp)
	move.l		d1,sypos(sp)
*-----------------------------------------------*
*	Clip right & bottom edges		*
*-----------------------------------------------*
	move.w		scrxsize,d1
	sub.w		d3,d1
	sub.w		d7,d1
	bpl.s		.xin
	add.w		d1,d7
	ble		.err
.xin:	move.w		scrysize,d1
	sub.w		d5,d1
	sub.w		d6,d1
	bpl.s		.yin
	add.w		d1,d6
	ble		.err
*-----------------------------------------------*
*	Clipped width & height now ready	*
*-----------------------------------------------*
.yin:	move.w		d7,cxsize(sp)
	move.w		d6,cysize(sp)
*-----------------------------------------------*
*	Address screen at dest X,Y		*
*-----------------------------------------------*
.mask:	mulu.w		scrxsize,d5
	add.l		d3,d5
	add.l		d5,d5
	add.l		d5,a5
*-----------------------------------------------*
*	Set up loop registers			*
*-----------------------------------------------*
	move.w		srcpwid,d7
	move.w		dstpwid,a6
	move.l		sypos(sp),d4
	move.l		syinc(sp),a3
	move.l		sxinc(sp),d2
	add.l		a6,a6
	move.w		d2,a2
	swap		d2
	moveq		#0,d3
	swap		d7
*-----------------------------------------------*
*	Create local variable structure		*
*-----------------------------------------------*
	bra		.vertical_start
*-----------------------------------------------*
*	Draw (height) scaled horizontal lines	*
*-----------------------------------------------*
.vertical_loop:
*-----------------------------------------------*
*	Vertical (outer) loop			*
*-----------------------------------------------*
	move.l		d4,d5
	move.w		sxpos(sp),d0
	sub.w		d5,d5
	swap		d5
	move.w		sxposf(sp),d1
	swap		d7
	lsl.l		d7,d5
	swap		d7
	lea		(a4,d5.l*2),a0
	move.l		a5,a1
*-----------------------------------------------*
*	Find codetower position & loopcount	*
*-----------------------------------------------*
	moveq		#8-1,d5
	and.w		d7,d5
	lsr.w		#3,d7
*-----------------------------------------------*
*	Method1 (bitwise) uses 22-byte module	*
*-----------------------------------------------*
	ifd		method1
	move.w		d5,d3
	lsl.w		#3,d5
	add.w		d3,d5
	add.w		d3,d5
	add.w		d3,d5
	neg.w		d5
	endc
*-----------------------------------------------*
*	Method2 (condition) uses 14-byte module	*
*-----------------------------------------------*
	ifd		method2
	move.w		d5,d3
	lsl.w		#3,d3
	sub.w		d3,d5
	endc
*-----------------------------------------------*
	jmp		(.base.w,pc,d5.w*2)
*-----------------------------------------------*
.horizontal_loop:
*-----------------------------------------------*
*	Expanded horizontal (inner) loop	*
*-----------------------------------------------*
	rept		8
	render
	endr
.base	dbra		d7,.horizontal_loop
*-----------------------------------------------*
*	Advance to next line (src & dest)	*
*-----------------------------------------------*
	add.l		a3,d4
	add.l		a6,a5
*-----------------------------------------------*
.vertical_start:
*-----------------------------------------------*
	move.w		cxsize(sp),d7
	dbra		d6,.vertical_loop
.err:	lea		scsp_size(sp),sp
	rts

dxsize:		ds.w	1
dysize:		ds.w	1
sxsize:		ds.w	1
sysize:		ds.w	1
dxpos:		ds.w	1
dypos:		ds.w	1
scrxsize:	ds.w	1
scrysize:	ds.w	1
srcpwid:	ds.w	1
dstpwid:	ds.w	1

*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*
*	System code				*
*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*

cls:	move.l		screen,a0
	move.l		#(screen_width*screen_height*2)/4,d0

clear_longs:
;	move.l		#$33333333,d2
	moveq		#0,d2
	move.l		d2,d3
	move.l		d2,d4
	move.l		d2,d5
	move.l		d2,d6
	move.l		d2,d7
	move.l		d2,a2
	move.l		d2,a3
	moveq		#128-1,d1
	and.l		d0,d1
	bra.s		.strt
.long	move.l		d2,(a0)+
.strt	dbra		d1,.long
	lsr.l		#7,d0
	move.l		d0,d1
	lsl.l		#8,d1				; *256
	lsl.l		#1,d1				; *2	=  *512
	add.l		d1,a0
	bra.s		.next
.clr	movem.l		d2-d7/a2-a3,-(a0)		; 4*8*16 = 512
	movem.l		d2-d7/a2-a3,-(a0)
	movem.l		d2-d7/a2-a3,-(a0)
	movem.l		d2-d7/a2-a3,-(a0)
	movem.l		d2-d7/a2-a3,-(a0)
	movem.l		d2-d7/a2-a3,-(a0)
	movem.l		d2-d7/a2-a3,-(a0)
	movem.l		d2-d7/a2-a3,-(a0)
	movem.l		d2-d7/a2-a3,-(a0)
	movem.l		d2-d7/a2-a3,-(a0)
	movem.l		d2-d7/a2-a3,-(a0)
	movem.l		d2-d7/a2-a3,-(a0)
	movem.l		d2-d7/a2-a3,-(a0)
	movem.l		d2-d7/a2-a3,-(a0)
	movem.l		d2-d7/a2-a3,-(a0)
	movem.l		d2-d7/a2-a3,-(a0)
.next	dbra		d0,.clr
	rts
	
set_screen:
	push.w		#2
	trap		#14
	addq		#2,sp
	move.l		d0,gem_screen
	push.w		#-1
	push.w		#88
	trap		#14
	addq		#4,sp
	move.w		d0,gem_video
	add.l		#255,screen
	and.l		#-256,screen
	push.w		#-1
	push.l		screen
	push.l		screen
	push.w		#5
	trap		#14
	lea		12(sp),sp
	push.w		#video_mode
	push.w		#88
	trap		#14
	addq.l		#4,sp
	rts

reset_screen:
	move.l		gem_screen,a0
	push.w		gem_video
	push.w		#-1
	push.l		a0
	push.l		a0
	push.w		#5
	trap		#14
	lea		12(sp),sp
	push.w		#88
	trap		#14
	addq		#4,sp
	rts

set_mouse:
	push.w		#34
	trap		#14
	addq		#2,sp
	move.l		d0,a0
	move.l		a0,old_kbd
	move.l		16(a0),old_mouse
	move.l		#new_mouse,16(a0)
	rts

reset_mouse:
	move.l		old_kbd,a0
	move.l		old_mouse,16(a0)
	rts

new_mouse:
	push.l		d0
	moveq		#4-1,d0
	and.b		(a0)+,d0
	lsr.w		d0
	bcc.s		.cc
	or.b		#2,d0
.cc	move.b		d0,buttons
	move.b		(a0)+,d0
	ext.w		d0
	tst.b		buttons
	bne.s		.xb
	add.w		d0,dxpos
	bra.s		.ok1	
.xb:	add.w		d0,dxsize
	bpl.s		.ok1
	clr.w		dxsize
.ok1:	move.b		(a0)+,d0
	ext.w		d0
	tst.b		buttons
	bne.s		.yb
	add.w		d0,dypos
	bra.s		.ok3
.yb:	add.w		d0,dysize
	bpl.s		.ok3
	clr.w		dysize
.ok3:	pop.l		d0
	rts

flush:	push.w		#11
	trap		#1
	addq		#2,sp
	tst.w		d0
	beq.s		.ot
	move.w		#7,-(sp)
	trap		#1
	addq		#2,sp
	bra.s		flush
.ot	rts

*-------------------------------------------------------------------------*
		data
*-------------------------------------------------------------------------*

screen:		dc.l	screen_1

*-----------------------------------------------------------------------*
*	Sprite data - 64 pixels wide / 60 pixels high			*
*-----------------------------------------------------------------------*
*	Only 40 pixels of width are used from the full 64, but the	*
*	source width must be rounded up to the next power of 2.		*
*-----------------------------------------------------------------------*

data:		incbin	l:\trooper.apx		; block  = 64 x 60
						; sprite = 40 x 60

*-----------------------------------------------------------------------*
		bss
*-----------------------------------------------------------------------*

*-----------------------------------------------------------------------*
*	General system space						*
*-----------------------------------------------------------------------*

		ds.l	512
sstack:		ds.l	1
		ds.l	512
ustack:		ds.l	1

gem_screen:	ds.l	1
gem_video:	ds.w	1

old_kbd:	ds.l	1
old_mouse:	ds.l	1
buttons:	ds.w	1

*-----------------------------------------------------------------------*
*	Buffer space							*
*-----------------------------------------------------------------------*

screen_1:	ds.w	(screen_width*256)+128
